#include "deep_zoom.h"
#include "display.h"

FRACTAL *current_fractal;
int resuming;

// save the current fractal to disk when control c is pressed

void cntrl_c_save(int sig)
{
  printf("Control-C\n");

  if(current_fractal!=NULL && current_fractal->options.fractal_filename!=NULL)
    {
      write_fractal_to_file(current_fractal, 
			    current_fractal->options.fractal_filename);
    }
  exit(0);
}

// seeds the random function with the time

void seed_rand()
{
  time_t t;

  t=time(NULL);
  srand((unsigned int) t);
}

// sets the dimensions of the fractal and allocate memory to store it

void fractal_init(FRACTAL *f, unsigned char incomplete, unsigned long completed_to)
{
  f->incomplete=incomplete;
  f->completed_to=completed_to;
  f->options.random=RANDOM;
}

void fractal_set_dimensions(FRACTAL *f, int w, int h)
{
  f->width=w;
  f->height=h;
} 


// sets the precision of the fractal

void fractal_set_precision(FRACTAL *f, unsigned long p)
{
  f->precision=p;
  f->cutoff=CUTOFF_VALUE;
}

void parse_args(int argc, char **argv, FRACTAL *ff)
{
  int c;

  // set defaults
  ff->options.image_filename="a.png";
  ff->options.fractal_filename=NULL;
  ff->options.zoom_level=ZOOM_LEVEL;

  while(1)
    {
      c=getopt(argc, argv, "f:i:h:w:c:z:rv");

      if(c==-1)
	break;

      switch(c)
	{
	case(':'):
	  printf("Missing parameter\n");
	  break;
	case('f'):
	  ff->options.fractal_filename=optarg;
	  break;
	case('i'):
	  ff->options.image_filename=optarg;
	  break;
	case('h'):
	  ff->height=atoi(optarg);
	  break;
	case('w'):
	  ff->width=atoi(optarg);
	  break;
	case('z'):
	  ff->options.zoom_level=atoi(optarg);
	  break;
	case('r'):
	  resuming=1;
	  break;
	case('c'):
	  ff->cutoff=atoi(optarg);
	  break;
	case('v'):
	  ff->options.random=TRUE;
	  break;
	}
    }
}

int main(int argc, char **argv)
{
  FRACTAL low_res, final_fractal;
  int i, q;
  mpf_t temp1, temp2;
  long final_precision;

  init_c0des();

  // current_fractal will be saved to disk when control C is pressed
  current_fractal=NULL;

  // initialize the dimensions of the final image and precisions of the 
  // low_res and final fractals
  
  fractal_set_dimensions(&final_fractal, WIDTH, HEIGHT);

  fractal_set_dimensions(&low_res, LOW_RES_WIDTH, LOW_RES_HEIGHT);
  fractal_set_precision(&low_res, LOW_RES_PRECISION);

  final_fractal.incomplete=1;
  final_fractal.completed_to=0;

  low_res.image_data=(DATA32 *) malloc(sizeof(DATA32) * low_res.width * low_res.height);

  resuming=0;
  seed_rand();
  init_garbage();

  // parse command line arguments

  parse_args(argc, argv, &final_fractal);

  final_precision=HIGH_RES_PRECISION+ZOOM_INCREMENT*final_fractal.options.zoom_level;

  fractal_set_precision(&final_fractal, final_precision);
  low_res.cutoff=final_fractal.cutoff;
  low_res.options.random=final_fractal.options.random;
  final_fractal.image_data=(DATA32 *) malloc(sizeof(DATA32) * final_fractal.width 
					     * final_fractal.height);
  

  // initialize the variables that control current coordinates/dimensions of
  // low resolution with fractal with the maximum precision that will
  // be needed for this fractal

  mpf_init2(low_res.top_left.re, final_precision);
  mpf_init2(low_res.top_left.im, final_precision);
  mpf_init2(low_res.bottom_right.re, final_precision);
  mpf_init2(low_res.bottom_right.im, final_precision);

  mpf_set_si(low_res.top_left.re, -2);
  mpf_set_si(low_res.top_left.im, 1);
  mpf_set_si(low_res.bottom_right.re, 1);
  mpf_set_si(low_res.bottom_right.im, -1);

  mpf_init2(temp1, final_precision);
  mpf_init2(temp2, final_precision);


  // hook the signal handler for control-c

  signal(SIGINT, cntrl_c_save);

/*    go though ZOOM_LEVEL iterations of calculating a 400 pixel square fractal of a  */
/*    region, dividing it into quarters, selecting a quarter that is the most active  */
/*    and then repeating the process on that quarter */
  

  if(resuming==0)
    {
      for(i=0;i<final_fractal.options.zoom_level;++i)
	{
	  printf("Zoom level: %d\n", i);

	  low_res.completed_to=0;
	  low_res.incomplete=1;

	  // the precision must gradually increase because in each step we are working with
	  // an area one quarter the size
	  low_res.precision+=ZOOM_INCREMENT;

	  // calculate a fractal for the dimensions and precision of low_res (fractal.c)
	  build_fractal(&low_res);
	  // select the quarter (indexed by integer q) that is most active (zoom_selection.c)
	  q=pick_quarter(&low_res);
	  printf("q=%d\n", q);

	  // adjust the coordinates of low_res fractal so that the next fractal is found using
	  // the coordinates selected in pick_quarter()

	  mpf_sub(temp1, low_res.bottom_right.re, low_res.top_left.re);
	  mpf_div_ui(temp1, temp1, 2);

	  switch(q)
	    {
	    case(0):
	      mpf_add(low_res.top_left.re, low_res.top_left.re, temp1);

	      mpf_sub(temp1, low_res.top_left.im, low_res.bottom_right.im);
	      mpf_div_ui(temp1, temp1, 2);
	      mpf_sub(low_res.bottom_right.im, low_res.top_left.im, temp1);
	      break;
	    case(1):
	      mpf_add(low_res.bottom_right.re, low_res.top_left.re, temp1);

	      mpf_sub(temp1, low_res.top_left.im, low_res.bottom_right.im);
	      mpf_div_ui(temp1, temp1, 2);
	      mpf_sub(low_res.bottom_right.im, low_res.top_left.im, temp1);
	      break;
	    case(2):
	      mpf_add(low_res.bottom_right.re, low_res.top_left.re, temp1);

	      mpf_sub(temp1, low_res.top_left.im, low_res.bottom_right.im);
	      mpf_div_ui(temp1, temp1, 2);
	      mpf_sub(low_res.top_left.im, low_res.top_left.im, temp1);
	      break;
	    case(3):	  
	      mpf_add(low_res.top_left.re, low_res.top_left.re, temp1);

	      mpf_sub(temp1, low_res.top_left.im, low_res.bottom_right.im);
	      mpf_div_ui(temp1, temp1, 2);
	      mpf_sub(low_res.top_left.im, low_res.top_left.im, temp1);
	      break;
	    }      
	}
    }
  free(low_res.image_data);

  current_fractal=&final_fractal;

  // generate a high-resolution fractal based on coordinates found by the zooming 
  // iterations and save it to a file

  printf("Rendering final image\n");
  final_fractal.top_left=low_res.top_left;
  final_fractal.bottom_right=low_res.bottom_right;

  if(resuming==1)
    {
      printf("Reading fractal from file\n");

      if(final_fractal.options.fractal_filename==NULL)
	{
	  printf("You must specify the data file to resume from\n");
	  exit(1);
	}
      read_fractal_from_file(&final_fractal, final_fractal.options.fractal_filename);

      printf("Top left real = ");
      mpf_out_str(stdout, 10, 20, final_fractal.top_left.re);
      printf("\n");
      printf("Top left imaginary = ");
      mpf_out_str(stdout, 10, 20, final_fractal.top_left.im);
      printf("\n");

      printf("Bottom right real = ");
      mpf_out_str(stdout, 10, 20, final_fractal.bottom_right.re);
      printf("\n");
      printf("Bottom right imaginary = ");
      mpf_out_str(stdout, 10, 20, final_fractal.bottom_right.im);
      printf("\n");

    }
  build_fractal(&final_fractal);

  printf("image=%s\n", final_fractal.options.image_filename);
  if(final_fractal.options.image_filename!=NULL)
    write_image_to_file(&final_fractal, final_fractal.options.image_filename);

  if(final_fractal.options.fractal_filename!=NULL)
    write_fractal_to_file(&final_fractal, final_fractal.options.fractal_filename);

  free(final_fractal.image_data);
  free_all_garbage();
  return(0);
}




