/* the functions to read and write fractal calculations to a file */

#include "deep_zoom.h"

// writes an mpf_t to the file as the overall information followed by the data

int write_mpf(int fd, mpf_t f)
{
  MPF_FILE_CHUNK h;

  h._mp_prec=f->_mp_prec;
  h._mp_size=f->_mp_size;
  h._mp_exp=f->_mp_exp;

  if(write(fd, &h, sizeof(MPF_FILE_CHUNK))!=sizeof(MPF_FILE_CHUNK))
    {
      return(1);
    }

  if(write(fd, f->_mp_d, sizeof(mp_limb_t)*absolute(f->_mp_size)) != 
     (signed) sizeof(mp_limb_t)*absolute(f->_mp_size) )
    return(1);
  return(0);
}

__mpf_struct read_mpf(int fd)
{
  __mpf_struct ret;
  MPF_FILE_CHUNK h;

  read(fd, &h, sizeof(MPF_FILE_CHUNK));

  ret._mp_prec=h._mp_prec;
  ret._mp_size=h._mp_size;                
  ret._mp_exp=h._mp_exp;

  ret._mp_d=(mp_limb_t *) malloc( sizeof(mp_limb_t) * absolute(ret._mp_size) );

  read(fd, ret._mp_d, sizeof(mp_limb_t) * absolute(ret._mp_size));
  return(ret);
}


// writes a string in the form of an int length followed by the characters of the string

int write_string_to_file(int fd, char *str)
{
  int len;

  len=strlen(str);

  if(write(fd, &len, sizeof(int))!=sizeof(int))
    {
      return(1);
    }

  if(write(fd, str, len)!=len)
    {
      return(2);
    }  
  return(0);
}

char *read_string_from_file(int fd)
{
  char *str;
  int len;

  if(read(fd, &len, sizeof(int))!=sizeof(int))
    {
      return( (char *) NULL);
    }

  if( (str= (char *) malloc(len)) == NULL)
    {
      return( (char *) NULL);
    }

  if(read(fd, str, len)!=len)
    {
      return( (char *) NULL);
    }
  str[len]='\0';
  return(str);
}

int write_fractal_to_file(FRACTAL *f, char *fname)
{
  FILE_FRACTAL ff;
  int fd;
  unsigned long bytes_to_write;

  if( (fd=open(fname, (O_WRONLY | O_TRUNC | O_CREAT), (S_IRUSR | S_IWUSR) ))==-1)
    {
      printf("fd=%d %s\n", fd, fname);
      return(1);
    }

  ff.incomplete=f->incomplete;
  ff.completed_to=f->completed_to;
  ff.height=f->height;
  ff.width=f->width;
  ff.precision=f->precision;
  ff.cutoff=f->cutoff;
  ff.zoom_level=f->options.zoom_level;
  ff.current_zoom_level=f->options.current_zoom_level;
  ff.random=f->options.random;

  // write the header chunk

  if(write(fd, &ff, sizeof(FILE_FRACTAL))!=sizeof(FILE_FRACTAL))
    {
      close(fd);
      return(2);
    }

  // write the filenames

  write_string_to_file(fd, f->options.image_filename);
  write_string_to_file(fd, f->options.fractal_filename);

  // write the mpf's
  
  if(write_mpf(fd, f->top_left.re)!=0)
    {
      close(fd);
      return(3);
    }

  if(write_mpf(fd, f->top_left.im)!=0)
    {
      close(fd);
      return(3);
    }

  if(write_mpf(fd, f->bottom_right.re)!=0)
    {
      close(fd);
      return(3);
    }

  if(write_mpf(fd, f->bottom_right.im)!=0)
    {
      close(fd);
      return(3);
    }

  // write the data

  if(f->incomplete)
    bytes_to_write=sizeof(DATA32) * f->completed_to;
  else
    bytes_to_write=sizeof(DATA32) * f->height * f->width;

  if(write(fd, f->image_data, bytes_to_write) != (signed) bytes_to_write)
    {
      close(fd);
      return(4);
    }
  close(fd);
  printf("Success\n");
  return(0);
}

int read_fractal_from_file(FRACTAL *f, char *fname)
{
  FILE_FRACTAL ff;
  int fd;
  unsigned long bytes_to_read=0;

  if( (fd=open(fname, O_RDONLY))==-1)
    {
      return(1);
    }

  // read the header

  if(read(fd, &ff, sizeof(FILE_FRACTAL))!=sizeof(FILE_FRACTAL))
    {
      return(2);
    }

  f->incomplete=ff.incomplete;
  f->completed_to=ff.completed_to;
  f->height=ff.height;
  f->width=ff.width;
  f->precision=ff.precision;
  f->cutoff=ff.cutoff;
  f->options.zoom_level=ff.zoom_level;
  f->options.current_zoom_level=ff.current_zoom_level;
  f->options.random=ff.random;

  // read filenames from file

  f->options.image_filename=read_string_from_file(fd);
  f->options.fractal_filename=read_string_from_file(fd);

  // read mpf's

  f->top_left.re[0]=read_mpf(fd);
  f->top_left.im[0]=read_mpf(fd);
  f->bottom_right.re[0]=read_mpf(fd);
  f->bottom_right.im[0]=read_mpf(fd);

  // read image data

  if(ff.incomplete)
    {
      bytes_to_read=ff.completed_to*sizeof(DATA32);
    }
  else
    {
      bytes_to_read=ff.width*ff.height*sizeof(DATA32);
    }

  // allocate enough memory for all future pixels
  f->image_data=(DATA32 *) malloc(ff.width*ff.height*sizeof(DATA32));
  read(fd, f->image_data, bytes_to_read);
  return(0);
}



