#include "deep_zoom.h"
#include "display.h"

// to be used to collect statistics about tree height

unsigned long n_values;
unsigned long total_height;
unsigned long n_nodes;

// an array of initialized mpf's to avoid having to initialize and clear them
mpf_t mpf_list[6];

// returns the number of iterations required for the value of (initial_x, initial_y) to escape

DATA32 escape_value(mpf_t initial_x, mpf_t initial_y, unsigned long cutoff)
{
  DATA32 value;
  //  DIGEST_TREE_NODE *periodicity_values;
  mpf_t x, y, temp1, temp2, x2, y2;
  COMPLEX_MPF period_entry_p;
  unsigned char does_not_escape=TRUE;

  // create a new binary search tree with a NULL first entry
  //periodicity_values=new_digest_node(NULL);
  init_lookup();

  /* for the purposes for calculating a fractal, 6 mpf's of the same size are 
     reused for each escape value. initing and clearing them only once in 
     build_fractal saves plenty of time (much less memory allocation) */

  x[0]=mpf_list[0][0];
  y[0]=mpf_list[1][0];
  temp1[0]=mpf_list[2][0];
  temp2[0]=mpf_list[3][0];
  x2[0]=mpf_list[4][0];
  y2[0]=mpf_list[5][0];

  for(value=0;value<cutoff;++value)
    {
      // calculate the absolute value of (x,y) (this is a complex number)
      mpf_pow_ui(x2, x, 2);
      mpf_pow_ui(y2, y, 2);
      mpf_add(temp1, x2, y2);

      // if this absolute value is greater than the escape value 
      // return the number of iterations needed

      if(mpf_cmp_ui(temp1, MAX_VALUE*MAX_VALUE)>0)
	{
	  does_not_escape=FALSE;
	  break;
	}
      period_entry_p.re[0]=x[0];
      period_entry_p.im[0]=y[0];

      // add the current (x, y) to a binary search tree of values to check if the current value has
      // occured before, if it has, a sequence of values will infinitely repeat without escaping

      if(add_to_periodicity_tree(&period_entry_p))
	{
	  set_fg(3);
	  set_bold_fg();
	  printf("%d ", value);
	  normal_scr();
	  fflush(stdout);
	  break;
	}

      // if the program gets to this point the last iteration did not escape, so the 
      // next iteration will be calculated

      mpf_sub(temp1, x2, y2);
      mpf_add(temp1, temp1, initial_x);

      mpf_mul(temp2, x, y);
      mpf_mul_ui(y, temp2, 2);
      mpf_add(y, y, initial_y);

      mpf_set(x, temp1);
    }

  free_all_trees();
  if(does_not_escape)
    return(DOES_NOT_ESCAPE_VALUE);

  return(value);
}

// sets the pixel in the image_data of f corresponding to (x, y) to d
// to avoid a tendency towards very soft blue images the data is shifted 8 bits and or'd with
// itself. Any modifications to the color scheme should be done here

void set_image_pt(FRACTAL *f, unsigned long x, unsigned long y, DATA32 d)
{
  DATA32 shifted;
  shifted=d<<8;
  shifted|=d;
  f->image_data[ (f->width*y) + x]=shifted;
}

// calculate the fractal specified by the dimensions and precision in f and put the resulting image
// into f->image_data

void build_fractal(FRACTAL *f)
{
  int i, j, height_i=0, width_i=0;
  mpf_t iterator_x, iterator_y, x_counter, y_counter, temp;

  n_values=0;
  n_nodes=0;
  total_height=0;

  // add mpf's initialized from here on will be of this precision
  mpf_set_default_prec(f->precision);

  for(i=0;i<6;++i)
    mpf_init(mpf_list[i]);

  if(f->incomplete)
    {
      height_i=f->completed_to/f->width;
      width_i=f->completed_to%f->width;
    }

  printf("Cutoff: %ld\n", f->cutoff);

  printf("Top left real = ");
  mpf_out_str(stdout, 10, 20, f->top_left.re);
  printf("\n");
  printf("Top left imaginary = ");
  mpf_out_str(stdout, 10, 20, f->top_left.im);
  printf("\n");

  printf("Bottom right real = ");
  mpf_out_str(stdout, 10, 20, f->bottom_right.re);
  printf("\n");
  printf("Bottom right imaginary = ");
  mpf_out_str(stdout, 10, 20, f->bottom_right.im);
  printf("\n");

  // iterator_x and iterator_y are the increment that the counters are increased by as each pixel of
  // the fractal is calculated

  mpf_init(iterator_x);
  mpf_init(iterator_y);

  mpf_sub(iterator_x, f->bottom_right.re, f->top_left.re);
  mpf_sub(iterator_y, f->top_left.im, f->bottom_right.im);

  mpf_div_ui(iterator_x, iterator_x, f->width);
  mpf_div_ui(iterator_y, iterator_y, f->height);

  // x_counter and y_counter store the current coordinates being calculated, so they are initialized
  // to the top_left value of the fractal

  mpf_init_set(x_counter, f->top_left.re);
  mpf_init_set(y_counter, f->top_left.im);

  mpf_init(temp);

  if(f->incomplete)
    {
      mpf_mul_ui(temp, iterator_x, width_i);
      mpf_add(x_counter, x_counter, temp);

      mpf_mul_ui(temp, iterator_y, height_i);
      mpf_sub(y_counter, y_counter, temp);
    }

  // calculate the escape value for each point of the fractal. The row currently being calculated
  // is printed to the screen

  printf("X counter = ");
  mpf_out_str(stdout, 10, 20, x_counter);
  printf("\n");
  printf("Y counter = ");
  mpf_out_str(stdout, 10, 20, y_counter);
  printf("\n");

  j=width_i;

  printf("Row: ");
  for(i=height_i;i<f->height;++i)
    {
	  set_fg(1);
	  printf("%d ", i);
	  normal_scr();
	  fflush(stdout);

	  while(j<f->width)
	    {
	      set_image_pt(f, j, i, escape_value(x_counter, y_counter, f->cutoff));
	      ++f->completed_to;
	      mpf_add(x_counter, x_counter, iterator_x);
	      ++j;
	    }
	  j=0;

	  mpf_set(x_counter, f->top_left.re);
	  mpf_sub(y_counter, y_counter, iterator_y);
    }

  // clear the mpf's used in this function

  for(i=0;i<6;++i)
    mpf_clear(mpf_list[i]);

  f->incomplete=0;
  mpf_clear(iterator_x);
  mpf_clear(iterator_y);
  mpf_clear(x_counter);
  mpf_clear(y_counter);
}






