#include "deep_zoom.h"
#include <string.h>

extern "C" {
#include "md4/global.h"
#include "md4/md4.h"
}


// garbage collection variables
static long nodes_collected;
static DIGEST_TREE_NODE *garbage; // singly linked list using right pointer, a FIFO stack

// periodicity lookup table

static DIGEST_TREE_NODE *period_lookup[256];
//DIGEST_TREE_NODE *period;

void init_garbage()
{
  nodes_collected=0;
  garbage=NULL;
}

void init_lookup()
{
  memset(period_lookup, 0, 256*sizeof(DIGEST_TREE_NODE *));
}

void free_all_trees()
{
  int i;
  
  for(i=0;i<256;++i)
    {
      if(period_lookup[i]!=NULL)
	{
	  free_digest_tree(period_lookup[i]);
	}
    }
}

// add the node to the garbage collection or free it

void garbage_free(DIGEST_TREE_NODE *t)
{
  // the garbage is full, free the node and return
  if(nodes_collected>=GARBAGE_SIZE || !GARBAGE_COLLECTION)
    {
      free(t);
      return;
    }

  // add the node to the garbage

  t->right=garbage;
  garbage=t;
  ++nodes_collected;
}

void free_all_garbage()
{
  DIGEST_TREE_NODE *it;

  while(garbage!=NULL)
    {  
      it=garbage;
      garbage=garbage->right;
      free(it);
    }
}

// recursively free the memory stored by the nodes of the periodicity binary search tree

void free_digest_tree(DIGEST_TREE_NODE *t)
{
  if(t==NULL)
    return;
  free_digest_tree(t->left);
  free_digest_tree(t->right);
  garbage_free(t);
}

// recursively find the height of the tree

long calc_height(DIGEST_TREE_NODE *t)
{
  long rh=0, lh=0;

  if(t->right!=NULL)
      rh=calc_height(t->right);
      
  if(t->left!=NULL)
      lh=calc_height(t->left);
  return ( 1 + ( (lh>rh) ? lh : rh));
}

// this function creates a new node for the periodicity tree the MD5 hash (data) as the key

DIGEST_TREE_NODE *new_digest_node(unsigned char *data)
{
  DIGEST_TREE_NODE *ret;

  if(nodes_collected>0 && GARBAGE_COLLECTION)
    {
      ret=garbage;
      garbage=garbage->right;
      --nodes_collected;
    }
  else
    {
      ret=(DIGEST_TREE_NODE *) malloc(sizeof(DIGEST_TREE_NODE));
    }

  ret->left=NULL;
  ret->right=NULL;
  if (data)
    memcpy( ret->digest, data, sizeof(unsigned char) * DIGEST_SIZE);
  else
    memset( ret->digest, 0, sizeof(unsigned char) * DIGEST_SIZE);
  return(ret);
}

// this function adds the MD5 hash of a mpf to the periodicity binary search tree
// if there is already a member with the same MD5 hash then the iterations are in an
// infinite loop because it has produced the same values already. In this case the
// function returns true and the program treats this pixel as one that will eventually
// repeat until the cutoff point (which it will)

static unsigned char find_or_add_to_digest_tree(DIGEST_TREE_NODE *t, unsigned char *d)
{
  int comp;

  // never happens =)
  //if(t==NULL)
  //  return(FALSE);

  /* 
   * special case for the root node moved to caller
   * [assumption made this is not called directly]
   */

  comp=memcmp(t->digest, d, sizeof(unsigned char) * DIGEST_SIZE);

  if(comp > 0)
  {
		  if(t->right!=NULL)
		  {
				  return(find_or_add_to_digest_tree(t->right, d));
		  }
		  t->right=new_digest_node(d);
		  return(FALSE);
  }

  if(comp < 0)
  {
		  if(t->left!=NULL)
		  {
				  return(find_or_add_to_digest_tree(t->left, d));
		  }
		  t->left=new_digest_node(d);
		  return(FALSE);
  }


  // implicit:
  // if(comp==0)
  return(TRUE);

}

unsigned char lookup_and_search(unsigned char *d)
{
  DIGEST_TREE_NODE *t;

  t=period_lookup[d[0]];

  if(t==NULL)
    {
      period_lookup[d[0]]=new_digest_node(d);
      return(FALSE);
    }
  return(find_or_add_to_digest_tree(t, d));
}

/* note: this will evaluate x twice, which doesn't matter in this case */
#define absolute(x) ( (x) < 0 ? -(x) : (x) )

// create an 128 bit hash from the bits of the two mpf's in p

//  unsigned char *generate_MD5_key(COMPLEX_MPF *p, unsigned char *digest)
//  {
//    MD5_CTX context;
 
//    memset(digest, 0, sizeof(unsigned char) * DIGEST_SIZE);

//    MD5Init(&context);
//    MD5Update(&context, (unsigned char *)p->im->_mp_d, absolute(p->im->_mp_size));
//    MD5Update(&context, (unsigned char *)p->re->_mp_d, absolute(p->re->_mp_size));
//    MD5Final(&context, digest);
//    return(digest);
//}

unsigned char *generate_MD4_key(COMPLEX_MPF *p, unsigned char *digest)
{
  MD4_CTX context;
 
  memset(digest, 0, sizeof(unsigned char) * DIGEST_SIZE);

  MD4Init(&context);
  MD4Update(&context, (unsigned char *)p->im->_mp_d, absolute(p->im->_mp_size));
  MD4Update(&context, (unsigned char *)p->re->_mp_d, absolute(p->re->_mp_size));
  MD4Final(&context, digest);
  return(digest);
}

// this function generates an MD5 key of a value and then adds it to the periodicity tree
// if the key already exists in the tree the function returns TRUE. This function is 
// called from fractal.c

unsigned char add_to_periodicity_tree(COMPLEX_MPF *p)
{
  unsigned char key[DIGEST_SIZE];

  generate_MD4_key(p, key);
  return(lookup_and_search(key));
}









